(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[6446],{10463:function(e,t,n){(window.__NEXT_P=window.__NEXT_P||[]).push(["/datastore-connection/[id]",function(){return n(88228)}])},88228:function(e,t,n){"use strict";n.r(t),n.d(t,{default:function(){return w}});var i=n(24246),l=n(1226),s=n(77181),a=n(86677),r=n(27378),o=n(16134),c=n(14047),u=n(77830),d=n(20682),f=n(94725),h=n(7617),x=n(58754),p=n(11274),v=n(26788),j=n(38212),y=n(50881),m=n(41337),_=()=>{let e=(0,o.T)(),{connection:t,connectionOption:n}=(0,o.C)(d.ZZ),[l,a]=(0,r.useState)(void 0),c=(0,r.useMemo)(()=>()=>{let e=[];return(null==l?void 0:l.options)&&l.options.forEach(n=>{let l;switch(n){case v.H5.CONNECTOR_PARAMETERS:l={label:n,content:(0,i.jsx)(p.s,{})};break;case v.H5.DATASET_CONFIGURATION:l=(null==t?void 0:t.key)?{label:n,content:(0,i.jsx)(j.Z,{})}:void 0;break;case v.H5.DSR_CUSTOMIZATION:l=(null==t?void 0:t.key)?{label:n,content:(0,i.jsx)(y.Z,{})}:void 0}l&&e.push(l)}),e},[null==t?void 0:t.key,null==l?void 0:l.options]);return(0,r.useEffect)(()=>{if(n){let t=v.IM.find(e=>e.type===(null==n?void 0:n.type));t&&(a(t),e((0,d.nj)(v.Ss[2])))}return()=>{e((0,d.mc)())}},[n,e]),t&&n?(0,i.jsxs)(i.Fragment,{children:[(0,i.jsx)(x.Z,{heading:"Connection manager",breadcrumbItems:[{title:"All connections",href:u.JR},{title:t.name}]}),(0,i.jsx)(s.X6q,{fontSize:"md",fontWeight:"semibold",maxHeight:"40px",mb:"4px",whiteSpace:"nowrap",children:(0,i.jsxs)(s.xuv,{alignItems:"center",display:"flex",children:[(0,i.jsx)(m.Z,{data:n}),(0,i.jsx)(s.xvT,{ml:"8px",children:t.name})]})}),(0,i.jsx)(s.gCW,{alignItems:"stretch",flex:"1",gap:"18px",children:(0,i.jsx)(h.Z,{data:c(),flexGrow:1,isLazy:!0})})]}):null},g=n(28120),w=()=>{let e=(0,o.T)(),t=(0,a.useRouter)(),{id:n}=t.query,{errorAlert:h}=(0,c.V)(),x=(0,o.C)(d.y_),[p,v]=(0,r.useState)(!0),j=(e,t)=>t.find(t=>{var n;return t.identifier===e.connection_type&&t.identifier!==g.Rj.SAAS||t.identifier===(null===(n=e.saas_config)||void 0===n?void 0:n.type)});return(0,r.useEffect)(()=>{let i=()=>{h("An error occurred while loading ".concat(t.query.id)),t.push(u.JR,void 0,{shallow:!0})},l=async t=>{try{let n=[];n.push(e(f.DM.endpoints.getDatastoreConnectionByKey.initiate(t))),0===x.length&&n.push(e(d.d8.endpoints.getAllConnectionTypes.initiate({search:""})));let l=await Promise.allSettled(n);if(l.every(e=>"fulfilled"===e.status)){e((0,d.lm)(l[0].value.data));let t=[...l[1]?l[1].value.data.items:x],n=j(l[0].value.data,t);e((0,d.yA)(n)),v(!1)}else i()}catch(e){i()}};return n&&l(n),()=>{}},[n]),(0,i.jsxs)(i.Fragment,{children:[p&&(0,i.jsx)(s.xuv,{display:"flex",h:"100vh",justifyContent:"center",children:(0,i.jsx)(s.M5Y,{children:(0,i.jsx)(s.$jN,{})})}),!p&&(0,i.jsx)(l.Z,{children:(0,i.jsx)(_,{})})]})}}},function(e){e.O(0,[6527,9327,9767,2888,9774,179],function(){return e(e.s=10463)}),_N_E=e.O()}]);