(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[5744],{24354:function(e,n,r){(window.__NEXT_P=window.__NEXT_P||[]).push(["/data-discovery/detection/[resourceUrn]",function(){return r(42705)}])},42705:function(e,n,r){"use strict";r.r(n);var t=r(24246);r(27378);var c=r(35287),i=r(77830),u=r(58754),s=r(69828),o=r(47182),a=r(1039);n.default=()=>{let{resourceUrn:e,navigateToDetectionResults:n}=(0,o.Z)();return(0,t.jsxs)(c.Z,{title:"Data detection",children:[(0,t.jsx)(u.Z,{heading:"Data detection",children:(0,t.jsx)(s.Z,{parentLink:i.XY,resourceUrn:e,onPathClick:e=>n({resourceUrn:e})})}),(0,t.jsx)(a.Z,{resourceUrn:e})]})}}},function(e){e.O(0,[8033,6060,4481,9999,6315,2397,2888,9774,179],function(){return e(e.s=24354)}),_N_E=e.O()}]);